import { inspect } from "../jsutils/inspect.mjs";

import { invariant } from "../jsutils/invariant.mjs";

import { isIterableObject } from "../jsutils/isIterableObject.mjs";

import { isObjectLike } from "../jsutils/isObjectLike.mjs";

import { isPromise } from "../jsutils/isPromise.mjs";

import { memoize3 } from "../jsutils/memoize3.mjs";

import { pathToArray, addPath } from "../jsutils/Path.mjs";

import { promiseForObject } from "../jsutils/promiseForObject.mjs";

import { promiseReduce } from "../jsutils/promiseReduce.mjs";

import { locatedError } from "../error/locatedError.mjs";

import { OperationTypeNode } from "../language/ast.mjs";

import { isNonNullType, isListType, isLeafType, isAbstractType, isObjectType } from "../type/definition.mjs";

import { SchemaMetaFieldDef, TypeMetaFieldDef, TypeNameMetaFieldDef } from "../type/introspection.mjs";

import { assertValidSchema } from "../type/validate.mjs";

import { collectFields, collectSubfields as collectSubfields$1 } from "./collectFields.mjs";

import { getVariableValues, getArgumentValues } from "./values.mjs";

import { e, GraphQLError } from "../shared.mjs";

const collectSubfields = memoize3(((exeContext, returnType, fieldNodes) => collectSubfields$1(exeContext.schema, exeContext.fragments, exeContext.variableValues, returnType, fieldNodes)));

function execute(args) {
  const {schema, rootValue} = args;
  assertValidExecutionArguments(schema);
  const exeContext = buildExecutionContext(args);
  if (!("schema" in exeContext)) {
    return {
      errors: exeContext
    };
  }
  try {
    const {operation} = exeContext;
    const result = function executeOperation(exeContext, operation, rootValue) {
      const rootType = exeContext.schema.getRootType(operation.operation);
      if (null == rootType) {
        throw new GraphQLError(`Schema is not configured to execute ${operation.operation} operation.`, {
          nodes: operation
        });
      }
      const rootFields = collectFields(exeContext.schema, exeContext.fragments, exeContext.variableValues, rootType, operation.selectionSet);
      const path = void 0;
      switch (operation.operation) {
       case OperationTypeNode.QUERY:
        return executeFields(exeContext, rootType, rootValue, path, rootFields);

       case OperationTypeNode.MUTATION:
        return function executeFieldsSerially(exeContext, parentType, sourceValue, path, fields) {
          return promiseReduce(fields.entries(), ((results, [responseName, fieldNodes]) => {
            const fieldPath = addPath(path, responseName, parentType.name);
            const result = executeField(exeContext, parentType, sourceValue, fieldNodes, fieldPath);
            if (void 0 === result) {
              return results;
            }
            if (isPromise(result)) {
              return result.then((resolvedResult => {
                results[responseName] = resolvedResult;
                return results;
              }));
            }
            results[responseName] = result;
            return results;
          }), Object.create(null));
        }(exeContext, rootType, rootValue, path, rootFields);

       case OperationTypeNode.SUBSCRIPTION:
        return executeFields(exeContext, rootType, rootValue, path, rootFields);
      }
    }(exeContext, operation, rootValue);
    if (isPromise(result)) {
      return result.then((data => buildResponse(data, exeContext.errors)), (error => {
        exeContext.errors.push(error);
        return buildResponse(null, exeContext.errors);
      }));
    }
    return buildResponse(result, exeContext.errors);
  } catch (error) {
    exeContext.errors.push(error);
    return buildResponse(null, exeContext.errors);
  }
}

function executeSync(args) {
  const result = execute(args);
  if (isPromise(result)) {
    throw new Error("GraphQL execution failed to complete synchronously.");
  }
  return result;
}

function buildResponse(data, errors) {
  return 0 === errors.length ? {
    data
  } : {
    errors,
    data
  };
}

function assertValidExecutionArguments(schema, document, rawVariableValues) {
  assertValidSchema(schema);
}

function buildExecutionContext(args) {
  var _definition$name, _operation$variableDe;
  const {schema, document, rootValue, contextValue, variableValues: rawVariableValues, operationName, fieldResolver, typeResolver, subscribeFieldResolver} = args;
  let operation;
  const fragments = Object.create(null);
  for (const definition of document.definitions) {
    switch (definition.kind) {
     case e.OPERATION_DEFINITION:
      if (null == operationName) {
        if (void 0 !== operation) {
          return [ new GraphQLError("Must provide operation name if query contains multiple operations.") ];
        }
        operation = definition;
      } else if ((null === (_definition$name = definition.name) || void 0 === _definition$name ? void 0 : _definition$name.value) === operationName) {
        operation = definition;
      }
      break;

     case e.FRAGMENT_DEFINITION:
      fragments[definition.name.value] = definition;
    }
  }
  if (!operation) {
    if (null != operationName) {
      return [ new GraphQLError(`Unknown operation named "${operationName}".`) ];
    }
    return [ new GraphQLError("Must provide an operation.") ];
  }
  const variableDefinitions = null !== (_operation$variableDe = operation.variableDefinitions) && void 0 !== _operation$variableDe ? _operation$variableDe : [];
  const coercedVariableValues = getVariableValues(schema, variableDefinitions, null != rawVariableValues ? rawVariableValues : {}, {
    maxErrors: 50
  });
  if (coercedVariableValues.errors) {
    return coercedVariableValues.errors;
  }
  return {
    schema,
    fragments,
    rootValue,
    contextValue,
    operation,
    variableValues: coercedVariableValues.coerced,
    fieldResolver: null != fieldResolver ? fieldResolver : defaultFieldResolver,
    typeResolver: null != typeResolver ? typeResolver : defaultTypeResolver,
    subscribeFieldResolver: null != subscribeFieldResolver ? subscribeFieldResolver : defaultFieldResolver,
    errors: []
  };
}

function executeFields(exeContext, parentType, sourceValue, path, fields) {
  const results = Object.create(null);
  let containsPromise = !1;
  for (const [responseName, fieldNodes] of fields.entries()) {
    const result = executeField(exeContext, parentType, sourceValue, fieldNodes, addPath(path, responseName, parentType.name));
    if (void 0 !== result) {
      results[responseName] = result;
      if (isPromise(result)) {
        containsPromise = !0;
      }
    }
  }
  if (!containsPromise) {
    return results;
  }
  return promiseForObject(results);
}

function executeField(exeContext, parentType, source, fieldNodes, path) {
  var _fieldDef$resolve;
  const fieldDef = getFieldDef(exeContext.schema, parentType, fieldNodes[0]);
  if (!fieldDef) {
    return;
  }
  const returnType = fieldDef.type;
  const resolveFn = null !== (_fieldDef$resolve = fieldDef.resolve) && void 0 !== _fieldDef$resolve ? _fieldDef$resolve : exeContext.fieldResolver;
  const info = buildResolveInfo(exeContext, fieldDef, fieldNodes, parentType, path);
  try {
    const args = getArgumentValues(fieldDef, fieldNodes[0], exeContext.variableValues);
    const result = resolveFn(source, args, exeContext.contextValue, info);
    let completed;
    if (isPromise(result)) {
      completed = result.then((resolved => completeValue(exeContext, returnType, fieldNodes, info, path, resolved)));
    } else {
      completed = completeValue(exeContext, returnType, fieldNodes, info, path, result);
    }
    if (isPromise(completed)) {
      return completed.then(void 0, (rawError => handleFieldError(locatedError(rawError, fieldNodes, pathToArray(path)), returnType, exeContext)));
    }
    return completed;
  } catch (rawError) {
    return handleFieldError(locatedError(rawError, fieldNodes, pathToArray(path)), returnType, exeContext);
  }
}

function buildResolveInfo(exeContext, fieldDef, fieldNodes, parentType, path) {
  return {
    fieldName: fieldDef.name,
    fieldNodes,
    returnType: fieldDef.type,
    parentType,
    path,
    schema: exeContext.schema,
    fragments: exeContext.fragments,
    rootValue: exeContext.rootValue,
    operation: exeContext.operation,
    variableValues: exeContext.variableValues
  };
}

function handleFieldError(error, returnType, exeContext) {
  if (isNonNullType(returnType)) {
    throw error;
  }
  exeContext.errors.push(error);
  return null;
}

function completeValue(exeContext, returnType, fieldNodes, info, path, result) {
  if (result instanceof Error) {
    throw result;
  }
  if (isNonNullType(returnType)) {
    const completed = completeValue(exeContext, returnType.ofType, fieldNodes, info, path, result);
    if (null === completed) {
      throw new Error(`Cannot return null for non-nullable field ${info.parentType.name}.${info.fieldName}.`);
    }
    return completed;
  }
  if (null == result) {
    return null;
  }
  if (isListType(returnType)) {
    return function completeListValue(exeContext, returnType, fieldNodes, info, path, result) {
      if (!isIterableObject(result)) {
        throw new GraphQLError(`Expected Iterable, but did not find one for field "${info.parentType.name}.${info.fieldName}".`);
      }
      const itemType = returnType.ofType;
      let containsPromise = !1;
      const completedResults = Array.from(result, ((item, index) => {
        const itemPath = addPath(path, index, void 0);
        try {
          let completedItem;
          if (isPromise(item)) {
            completedItem = item.then((resolved => completeValue(exeContext, itemType, fieldNodes, info, itemPath, resolved)));
          } else {
            completedItem = completeValue(exeContext, itemType, fieldNodes, info, itemPath, item);
          }
          if (isPromise(completedItem)) {
            containsPromise = !0;
            return completedItem.then(void 0, (rawError => handleFieldError(locatedError(rawError, fieldNodes, pathToArray(itemPath)), itemType, exeContext)));
          }
          return completedItem;
        } catch (rawError) {
          return handleFieldError(locatedError(rawError, fieldNodes, pathToArray(itemPath)), itemType, exeContext);
        }
      }));
      return containsPromise ? Promise.all(completedResults) : completedResults;
    }(exeContext, returnType, fieldNodes, info, path, result);
  }
  if (isLeafType(returnType)) {
    return function completeLeafValue(returnType, result) {
      const serializedResult = returnType.serialize(result);
      if (null == serializedResult) {
        throw new Error(`Expected \`${inspect(returnType)}.serialize(${inspect(result)})\` to return non-nullable value, returned: ${inspect(serializedResult)}`);
      }
      return serializedResult;
    }(returnType, result);
  }
  if (isAbstractType(returnType)) {
    return function completeAbstractValue(exeContext, returnType, fieldNodes, info, path, result) {
      var _returnType$resolveTy;
      const resolveTypeFn = null !== (_returnType$resolveTy = returnType.resolveType) && void 0 !== _returnType$resolveTy ? _returnType$resolveTy : exeContext.typeResolver;
      const contextValue = exeContext.contextValue;
      const runtimeType = resolveTypeFn(result, contextValue, info, returnType);
      if (isPromise(runtimeType)) {
        return runtimeType.then((resolvedRuntimeType => completeObjectValue(exeContext, ensureValidRuntimeType(resolvedRuntimeType, exeContext, returnType, fieldNodes, info, result), fieldNodes, info, path, result)));
      }
      return completeObjectValue(exeContext, ensureValidRuntimeType(runtimeType, exeContext, returnType, fieldNodes, info, result), fieldNodes, info, path, result);
    }(exeContext, returnType, fieldNodes, info, path, result);
  }
  if (isObjectType(returnType)) {
    return completeObjectValue(exeContext, returnType, fieldNodes, info, path, result);
  }
  invariant(!1);
}

function ensureValidRuntimeType(runtimeTypeName, exeContext, returnType, fieldNodes, info, result) {
  if (null == runtimeTypeName) {
    throw new GraphQLError(`Abstract type "${returnType.name}" must resolve to an Object type at runtime for field "${info.parentType.name}.${info.fieldName}". Either the "${returnType.name}" type should provide a "resolveType" function or each possible type should provide an "isTypeOf" function.`, fieldNodes);
  }
  if (isObjectType(runtimeTypeName)) {
    throw new GraphQLError("Support for returning GraphQLObjectType from resolveType was removed in graphql-js@16.0.0 please return type name instead.");
  }
  if ("string" != typeof runtimeTypeName) {
    throw new GraphQLError(`Abstract type "${returnType.name}" must resolve to an Object type at runtime for field "${info.parentType.name}.${info.fieldName}" with value ${inspect(result)}, received "${inspect(runtimeTypeName)}".`);
  }
  const runtimeType = exeContext.schema.getType(runtimeTypeName);
  if (null == runtimeType) {
    throw new GraphQLError(`Abstract type "${returnType.name}" was resolved to a type "${runtimeTypeName}" that does not exist inside the schema.`, {
      nodes: fieldNodes
    });
  }
  if (!isObjectType(runtimeType)) {
    throw new GraphQLError(`Abstract type "${returnType.name}" was resolved to a non-object type "${runtimeTypeName}".`, {
      nodes: fieldNodes
    });
  }
  if (!exeContext.schema.isSubType(returnType, runtimeType)) {
    throw new GraphQLError(`Runtime Object type "${runtimeType.name}" is not a possible type for "${returnType.name}".`, {
      nodes: fieldNodes
    });
  }
  return runtimeType;
}

function completeObjectValue(exeContext, returnType, fieldNodes, info, path, result) {
  const subFieldNodes = collectSubfields(exeContext, returnType, fieldNodes);
  if (returnType.isTypeOf) {
    const isTypeOf = returnType.isTypeOf(result, exeContext.contextValue, info);
    if (isPromise(isTypeOf)) {
      return isTypeOf.then((resolvedIsTypeOf => {
        if (!resolvedIsTypeOf) {
          throw invalidReturnTypeError(returnType, result, fieldNodes);
        }
        return executeFields(exeContext, returnType, result, path, subFieldNodes);
      }));
    }
    if (!isTypeOf) {
      throw invalidReturnTypeError(returnType, result, fieldNodes);
    }
  }
  return executeFields(exeContext, returnType, result, path, subFieldNodes);
}

function invalidReturnTypeError(returnType, result, fieldNodes) {
  return new GraphQLError(`Expected value of type "${returnType.name}" but got: ${inspect(result)}.`, {
    nodes: fieldNodes
  });
}

const defaultTypeResolver = function(value, contextValue, info, abstractType) {
  if (isObjectLike(value) && "string" == typeof value.__typename) {
    return value.__typename;
  }
  const possibleTypes = info.schema.getPossibleTypes(abstractType);
  const promisedIsTypeOfResults = [];
  for (let i = 0; i < possibleTypes.length; i++) {
    const type = possibleTypes[i];
    if (type.isTypeOf) {
      const isTypeOfResult = type.isTypeOf(value, contextValue, info);
      if (isPromise(isTypeOfResult)) {
        promisedIsTypeOfResults[i] = isTypeOfResult;
      } else if (isTypeOfResult) {
        return type.name;
      }
    }
  }
  if (promisedIsTypeOfResults.length) {
    return Promise.all(promisedIsTypeOfResults).then((isTypeOfResults => {
      for (let i = 0; i < isTypeOfResults.length; i++) {
        if (isTypeOfResults[i]) {
          return possibleTypes[i].name;
        }
      }
    }));
  }
};

const defaultFieldResolver = function(source, args, contextValue, info) {
  if (isObjectLike(source) || "function" == typeof source) {
    const property = source[info.fieldName];
    if ("function" == typeof property) {
      return source[info.fieldName](args, contextValue, info);
    }
    return property;
  }
};

function getFieldDef(schema, parentType, fieldNode) {
  const fieldName = fieldNode.name.value;
  if (fieldName === SchemaMetaFieldDef.name && schema.getQueryType() === parentType) {
    return SchemaMetaFieldDef;
  } else if (fieldName === TypeMetaFieldDef.name && schema.getQueryType() === parentType) {
    return TypeMetaFieldDef;
  } else if (fieldName === TypeNameMetaFieldDef.name) {
    return TypeNameMetaFieldDef;
  }
  return parentType.getFields()[fieldName];
}

export { assertValidExecutionArguments, buildExecutionContext, buildResolveInfo, defaultFieldResolver, defaultTypeResolver, execute, executeSync, getFieldDef };
