var jsutils_invariant = require("../jsutils/invariant.js");

var jsutils_keyMap = require("../jsutils/keyMap.js");

var type_definition = require("../type/definition.js");

var type_scalars = require("../type/scalars.js");

var utilities_astFromValue = require("./astFromValue.js");

var utilities_sortValueNode = require("./sortValueNode.js");

var shared = require("../shared.js");

exports.BreakingChangeType = void 0;

!function(BreakingChangeType) {
  BreakingChangeType.TYPE_REMOVED = "TYPE_REMOVED";
  BreakingChangeType.TYPE_CHANGED_KIND = "TYPE_CHANGED_KIND";
  BreakingChangeType.TYPE_REMOVED_FROM_UNION = "TYPE_REMOVED_FROM_UNION";
  BreakingChangeType.VALUE_REMOVED_FROM_ENUM = "VALUE_REMOVED_FROM_ENUM";
  BreakingChangeType.REQUIRED_INPUT_FIELD_ADDED = "REQUIRED_INPUT_FIELD_ADDED";
  BreakingChangeType.IMPLEMENTED_INTERFACE_REMOVED = "IMPLEMENTED_INTERFACE_REMOVED";
  BreakingChangeType.FIELD_REMOVED = "FIELD_REMOVED";
  BreakingChangeType.FIELD_CHANGED_KIND = "FIELD_CHANGED_KIND";
  BreakingChangeType.REQUIRED_ARG_ADDED = "REQUIRED_ARG_ADDED";
  BreakingChangeType.ARG_REMOVED = "ARG_REMOVED";
  BreakingChangeType.ARG_CHANGED_KIND = "ARG_CHANGED_KIND";
  BreakingChangeType.DIRECTIVE_REMOVED = "DIRECTIVE_REMOVED";
  BreakingChangeType.DIRECTIVE_ARG_REMOVED = "DIRECTIVE_ARG_REMOVED";
  BreakingChangeType.REQUIRED_DIRECTIVE_ARG_ADDED = "REQUIRED_DIRECTIVE_ARG_ADDED";
  BreakingChangeType.DIRECTIVE_REPEATABLE_REMOVED = "DIRECTIVE_REPEATABLE_REMOVED";
  BreakingChangeType.DIRECTIVE_LOCATION_REMOVED = "DIRECTIVE_LOCATION_REMOVED";
}(exports.BreakingChangeType || (exports.BreakingChangeType = {}));

exports.DangerousChangeType = void 0;

!function(DangerousChangeType) {
  DangerousChangeType.VALUE_ADDED_TO_ENUM = "VALUE_ADDED_TO_ENUM";
  DangerousChangeType.TYPE_ADDED_TO_UNION = "TYPE_ADDED_TO_UNION";
  DangerousChangeType.OPTIONAL_INPUT_FIELD_ADDED = "OPTIONAL_INPUT_FIELD_ADDED";
  DangerousChangeType.OPTIONAL_ARG_ADDED = "OPTIONAL_ARG_ADDED";
  DangerousChangeType.IMPLEMENTED_INTERFACE_ADDED = "IMPLEMENTED_INTERFACE_ADDED";
  DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE = "ARG_DEFAULT_VALUE_CHANGE";
}(exports.DangerousChangeType || (exports.DangerousChangeType = {}));

function findSchemaChanges(oldSchema, newSchema) {
  return [ ...findTypeChanges(oldSchema, newSchema), ...findDirectiveChanges(oldSchema, newSchema) ];
}

function findDirectiveChanges(oldSchema, newSchema) {
  const schemaChanges = [];
  const directivesDiff = diff(oldSchema.getDirectives(), newSchema.getDirectives());
  for (const oldDirective of directivesDiff.removed) {
    schemaChanges.push({
      type: exports.BreakingChangeType.DIRECTIVE_REMOVED,
      description: `${oldDirective.name} was removed.`
    });
  }
  for (const [oldDirective, newDirective] of directivesDiff.persisted) {
    const argsDiff = diff(oldDirective.args, newDirective.args);
    for (const newArg of argsDiff.added) {
      if (type_definition.isRequiredArgument(newArg)) {
        schemaChanges.push({
          type: exports.BreakingChangeType.REQUIRED_DIRECTIVE_ARG_ADDED,
          description: `A required arg ${newArg.name} on directive ${oldDirective.name} was added.`
        });
      }
    }
    for (const oldArg of argsDiff.removed) {
      schemaChanges.push({
        type: exports.BreakingChangeType.DIRECTIVE_ARG_REMOVED,
        description: `${oldArg.name} was removed from ${oldDirective.name}.`
      });
    }
    if (oldDirective.isRepeatable && !newDirective.isRepeatable) {
      schemaChanges.push({
        type: exports.BreakingChangeType.DIRECTIVE_REPEATABLE_REMOVED,
        description: `Repeatable flag was removed from ${oldDirective.name}.`
      });
    }
    for (const location of oldDirective.locations) {
      if (!newDirective.locations.includes(location)) {
        schemaChanges.push({
          type: exports.BreakingChangeType.DIRECTIVE_LOCATION_REMOVED,
          description: `${location} was removed from ${oldDirective.name}.`
        });
      }
    }
  }
  return schemaChanges;
}

function findTypeChanges(oldSchema, newSchema) {
  const schemaChanges = [];
  const typesDiff = diff(Object.values(oldSchema.getTypeMap()), Object.values(newSchema.getTypeMap()));
  for (const oldType of typesDiff.removed) {
    schemaChanges.push({
      type: exports.BreakingChangeType.TYPE_REMOVED,
      description: type_scalars.isSpecifiedScalarType(oldType) ? `Standard scalar ${oldType.name} was removed because it is not referenced anymore.` : `${oldType.name} was removed.`
    });
  }
  for (const [oldType, newType] of typesDiff.persisted) {
    if (type_definition.isEnumType(oldType) && type_definition.isEnumType(newType)) {
      schemaChanges.push(...findEnumTypeChanges(oldType, newType));
    } else if (type_definition.isUnionType(oldType) && type_definition.isUnionType(newType)) {
      schemaChanges.push(...findUnionTypeChanges(oldType, newType));
    } else if (type_definition.isInputObjectType(oldType) && type_definition.isInputObjectType(newType)) {
      schemaChanges.push(...findInputObjectTypeChanges(oldType, newType));
    } else if (type_definition.isObjectType(oldType) && type_definition.isObjectType(newType)) {
      schemaChanges.push(...findFieldChanges(oldType, newType), ...findImplementedInterfacesChanges(oldType, newType));
    } else if (type_definition.isInterfaceType(oldType) && type_definition.isInterfaceType(newType)) {
      schemaChanges.push(...findFieldChanges(oldType, newType), ...findImplementedInterfacesChanges(oldType, newType));
    } else if (oldType.constructor !== newType.constructor) {
      schemaChanges.push({
        type: exports.BreakingChangeType.TYPE_CHANGED_KIND,
        description: `${oldType.name} changed from ${typeKindName(oldType)} to ${typeKindName(newType)}.`
      });
    }
  }
  return schemaChanges;
}

function findInputObjectTypeChanges(oldType, newType) {
  const schemaChanges = [];
  const fieldsDiff = diff(Object.values(oldType.getFields()), Object.values(newType.getFields()));
  for (const newField of fieldsDiff.added) {
    if (type_definition.isRequiredInputField(newField)) {
      schemaChanges.push({
        type: exports.BreakingChangeType.REQUIRED_INPUT_FIELD_ADDED,
        description: `A required field ${newField.name} on input type ${oldType.name} was added.`
      });
    } else {
      schemaChanges.push({
        type: exports.DangerousChangeType.OPTIONAL_INPUT_FIELD_ADDED,
        description: `An optional field ${newField.name} on input type ${oldType.name} was added.`
      });
    }
  }
  for (const oldField of fieldsDiff.removed) {
    schemaChanges.push({
      type: exports.BreakingChangeType.FIELD_REMOVED,
      description: `${oldType.name}.${oldField.name} was removed.`
    });
  }
  for (const [oldField, newField] of fieldsDiff.persisted) {
    if (!isChangeSafeForInputObjectFieldOrFieldArg(oldField.type, newField.type)) {
      schemaChanges.push({
        type: exports.BreakingChangeType.FIELD_CHANGED_KIND,
        description: `${oldType.name}.${oldField.name} changed type from ${String(oldField.type)} to ${String(newField.type)}.`
      });
    }
  }
  return schemaChanges;
}

function findUnionTypeChanges(oldType, newType) {
  const schemaChanges = [];
  const possibleTypesDiff = diff(oldType.getTypes(), newType.getTypes());
  for (const newPossibleType of possibleTypesDiff.added) {
    schemaChanges.push({
      type: exports.DangerousChangeType.TYPE_ADDED_TO_UNION,
      description: `${newPossibleType.name} was added to union type ${oldType.name}.`
    });
  }
  for (const oldPossibleType of possibleTypesDiff.removed) {
    schemaChanges.push({
      type: exports.BreakingChangeType.TYPE_REMOVED_FROM_UNION,
      description: `${oldPossibleType.name} was removed from union type ${oldType.name}.`
    });
  }
  return schemaChanges;
}

function findEnumTypeChanges(oldType, newType) {
  const schemaChanges = [];
  const valuesDiff = diff(oldType.getValues(), newType.getValues());
  for (const newValue of valuesDiff.added) {
    schemaChanges.push({
      type: exports.DangerousChangeType.VALUE_ADDED_TO_ENUM,
      description: `${newValue.name} was added to enum type ${oldType.name}.`
    });
  }
  for (const oldValue of valuesDiff.removed) {
    schemaChanges.push({
      type: exports.BreakingChangeType.VALUE_REMOVED_FROM_ENUM,
      description: `${oldValue.name} was removed from enum type ${oldType.name}.`
    });
  }
  return schemaChanges;
}

function findImplementedInterfacesChanges(oldType, newType) {
  const schemaChanges = [];
  const interfacesDiff = diff(oldType.getInterfaces(), newType.getInterfaces());
  for (const newInterface of interfacesDiff.added) {
    schemaChanges.push({
      type: exports.DangerousChangeType.IMPLEMENTED_INTERFACE_ADDED,
      description: `${newInterface.name} added to interfaces implemented by ${oldType.name}.`
    });
  }
  for (const oldInterface of interfacesDiff.removed) {
    schemaChanges.push({
      type: exports.BreakingChangeType.IMPLEMENTED_INTERFACE_REMOVED,
      description: `${oldType.name} no longer implements interface ${oldInterface.name}.`
    });
  }
  return schemaChanges;
}

function findFieldChanges(oldType, newType) {
  const schemaChanges = [];
  const fieldsDiff = diff(Object.values(oldType.getFields()), Object.values(newType.getFields()));
  for (const oldField of fieldsDiff.removed) {
    schemaChanges.push({
      type: exports.BreakingChangeType.FIELD_REMOVED,
      description: `${oldType.name}.${oldField.name} was removed.`
    });
  }
  for (const [oldField, newField] of fieldsDiff.persisted) {
    schemaChanges.push(...findArgChanges(oldType, oldField, newField));
    if (!isChangeSafeForObjectOrInterfaceField(oldField.type, newField.type)) {
      schemaChanges.push({
        type: exports.BreakingChangeType.FIELD_CHANGED_KIND,
        description: `${oldType.name}.${oldField.name} changed type from ${String(oldField.type)} to ${String(newField.type)}.`
      });
    }
  }
  return schemaChanges;
}

function findArgChanges(oldType, oldField, newField) {
  const schemaChanges = [];
  const argsDiff = diff(oldField.args, newField.args);
  for (const oldArg of argsDiff.removed) {
    schemaChanges.push({
      type: exports.BreakingChangeType.ARG_REMOVED,
      description: `${oldType.name}.${oldField.name} arg ${oldArg.name} was removed.`
    });
  }
  for (const [oldArg, newArg] of argsDiff.persisted) {
    if (!isChangeSafeForInputObjectFieldOrFieldArg(oldArg.type, newArg.type)) {
      schemaChanges.push({
        type: exports.BreakingChangeType.ARG_CHANGED_KIND,
        description: `${oldType.name}.${oldField.name} arg ${oldArg.name} has changed type from ${String(oldArg.type)} to ${String(newArg.type)}.`
      });
    } else if (void 0 !== oldArg.defaultValue) {
      if (void 0 === newArg.defaultValue) {
        schemaChanges.push({
          type: exports.DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
          description: `${oldType.name}.${oldField.name} arg ${oldArg.name} defaultValue was removed.`
        });
      } else {
        const oldValueStr = stringifyValue(oldArg.defaultValue, oldArg.type);
        const newValueStr = stringifyValue(newArg.defaultValue, newArg.type);
        if (oldValueStr !== newValueStr) {
          schemaChanges.push({
            type: exports.DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
            description: `${oldType.name}.${oldField.name} arg ${oldArg.name} has changed defaultValue from ${oldValueStr} to ${newValueStr}.`
          });
        }
      }
    }
  }
  for (const newArg of argsDiff.added) {
    if (type_definition.isRequiredArgument(newArg)) {
      schemaChanges.push({
        type: exports.BreakingChangeType.REQUIRED_ARG_ADDED,
        description: `A required arg ${newArg.name} on ${oldType.name}.${oldField.name} was added.`
      });
    } else {
      schemaChanges.push({
        type: exports.DangerousChangeType.OPTIONAL_ARG_ADDED,
        description: `An optional arg ${newArg.name} on ${oldType.name}.${oldField.name} was added.`
      });
    }
  }
  return schemaChanges;
}

function isChangeSafeForObjectOrInterfaceField(oldType, newType) {
  if (type_definition.isListType(oldType)) {
    return type_definition.isListType(newType) && isChangeSafeForObjectOrInterfaceField(oldType.ofType, newType.ofType) || type_definition.isNonNullType(newType) && isChangeSafeForObjectOrInterfaceField(oldType, newType.ofType);
  }
  if (type_definition.isNonNullType(oldType)) {
    return type_definition.isNonNullType(newType) && isChangeSafeForObjectOrInterfaceField(oldType.ofType, newType.ofType);
  }
  return type_definition.isNamedType(newType) && oldType.name === newType.name || type_definition.isNonNullType(newType) && isChangeSafeForObjectOrInterfaceField(oldType, newType.ofType);
}

function isChangeSafeForInputObjectFieldOrFieldArg(oldType, newType) {
  if (type_definition.isListType(oldType)) {
    return type_definition.isListType(newType) && isChangeSafeForInputObjectFieldOrFieldArg(oldType.ofType, newType.ofType);
  }
  if (type_definition.isNonNullType(oldType)) {
    return type_definition.isNonNullType(newType) && isChangeSafeForInputObjectFieldOrFieldArg(oldType.ofType, newType.ofType) || !type_definition.isNonNullType(newType) && isChangeSafeForInputObjectFieldOrFieldArg(oldType.ofType, newType);
  }
  return type_definition.isNamedType(newType) && oldType.name === newType.name;
}

function typeKindName(type) {
  if (type_definition.isScalarType(type)) {
    return "a Scalar type";
  }
  if (type_definition.isObjectType(type)) {
    return "an Object type";
  }
  if (type_definition.isInterfaceType(type)) {
    return "an Interface type";
  }
  if (type_definition.isUnionType(type)) {
    return "a Union type";
  }
  if (type_definition.isEnumType(type)) {
    return "an Enum type";
  }
  if (type_definition.isInputObjectType(type)) {
    return "an Input type";
  }
  jsutils_invariant.invariant(!1);
}

function stringifyValue(value, type) {
  const ast = utilities_astFromValue.astFromValue(value, type);
  null != ast || jsutils_invariant.invariant(!1);
  return shared.print(utilities_sortValueNode.sortValueNode(ast));
}

function diff(oldArray, newArray) {
  const added = [];
  const removed = [];
  const persisted = [];
  const oldMap = jsutils_keyMap.keyMap(oldArray, (({name}) => name));
  const newMap = jsutils_keyMap.keyMap(newArray, (({name}) => name));
  for (const oldItem of oldArray) {
    const newItem = newMap[oldItem.name];
    if (void 0 === newItem) {
      removed.push(oldItem);
    } else {
      persisted.push([ oldItem, newItem ]);
    }
  }
  for (const newItem of newArray) {
    if (void 0 === oldMap[newItem.name]) {
      added.push(newItem);
    }
  }
  return {
    added,
    persisted,
    removed
  };
}

exports.findBreakingChanges = function findBreakingChanges(oldSchema, newSchema) {
  return findSchemaChanges(oldSchema, newSchema).filter((change => change.type in exports.BreakingChangeType));
};

exports.findDangerousChanges = function findDangerousChanges(oldSchema, newSchema) {
  return findSchemaChanges(oldSchema, newSchema).filter((change => change.type in exports.DangerousChangeType));
};
