import { invariant } from "../jsutils/invariant.mjs";

import { keyValMap } from "../jsutils/keyValMap.mjs";

import { naturalCompare } from "../jsutils/naturalCompare.mjs";

import { isScalarType, isObjectType, GraphQLObjectType, isInterfaceType, GraphQLInterfaceType, isUnionType, GraphQLUnionType, isEnumType, GraphQLEnumType, isInputObjectType, GraphQLInputObjectType, isListType, GraphQLList, isNonNullType, GraphQLNonNull } from "../type/definition.mjs";

import { GraphQLDirective } from "../type/directives.mjs";

import { isIntrospectionType } from "../type/introspection.mjs";

import { GraphQLSchema } from "../type/schema.mjs";

function lexicographicSortSchema(schema) {
  const schemaConfig = schema.toConfig();
  const typeMap = keyValMap(sortByName(schemaConfig.types), (type => type.name), (function sortNamedType(type) {
    if (isScalarType(type) || isIntrospectionType(type)) {
      return type;
    }
    if (isObjectType(type)) {
      const config = type.toConfig();
      return new GraphQLObjectType({
        ...config,
        interfaces: () => sortTypes(config.interfaces),
        fields: () => sortFields(config.fields)
      });
    }
    if (isInterfaceType(type)) {
      const config = type.toConfig();
      return new GraphQLInterfaceType({
        ...config,
        interfaces: () => sortTypes(config.interfaces),
        fields: () => sortFields(config.fields)
      });
    }
    if (isUnionType(type)) {
      const config = type.toConfig();
      return new GraphQLUnionType({
        ...config,
        types: () => sortTypes(config.types)
      });
    }
    if (isEnumType(type)) {
      const config = type.toConfig();
      return new GraphQLEnumType({
        ...config,
        values: sortObjMap(config.values, (value => value))
      });
    }
    if (isInputObjectType(type)) {
      const config = type.toConfig();
      return new GraphQLInputObjectType({
        ...config,
        fields: () => function sortInputFields(fieldsMap) {
          return sortObjMap(fieldsMap, (field => ({
            ...field,
            type: replaceType(field.type)
          })));
        }(config.fields)
      });
    }
    invariant(!1);
  }));
  return new GraphQLSchema({
    ...schemaConfig,
    types: Object.values(typeMap),
    directives: sortByName(schemaConfig.directives).map((function sortDirective(directive) {
      const config = directive.toConfig();
      return new GraphQLDirective({
        ...config,
        locations: sortBy(config.locations, (x => x)),
        args: sortArgs(config.args)
      });
    })),
    query: replaceMaybeType(schemaConfig.query),
    mutation: replaceMaybeType(schemaConfig.mutation),
    subscription: replaceMaybeType(schemaConfig.subscription)
  });
  function replaceType(type) {
    if (isListType(type)) {
      return new GraphQLList(replaceType(type.ofType));
    } else if (isNonNullType(type)) {
      return new GraphQLNonNull(replaceType(type.ofType));
    }
    return replaceNamedType(type);
  }
  function replaceNamedType(type) {
    return typeMap[type.name];
  }
  function replaceMaybeType(maybeType) {
    return maybeType && replaceNamedType(maybeType);
  }
  function sortArgs(args) {
    return sortObjMap(args, (arg => ({
      ...arg,
      type: replaceType(arg.type)
    })));
  }
  function sortFields(fieldsMap) {
    return sortObjMap(fieldsMap, (field => ({
      ...field,
      type: replaceType(field.type),
      args: field.args && sortArgs(field.args)
    })));
  }
  function sortTypes(array) {
    return sortByName(array).map(replaceNamedType);
  }
}

function sortObjMap(map, sortValueFn) {
  const sortedMap = Object.create(null);
  for (const key of Object.keys(map).sort(naturalCompare)) {
    sortedMap[key] = sortValueFn(map[key]);
  }
  return sortedMap;
}

function sortByName(array) {
  return sortBy(array, (obj => obj.name));
}

function sortBy(array, mapToKey) {
  return array.slice().sort(((obj1, obj2) => {
    const key1 = mapToKey(obj1);
    const key2 = mapToKey(obj2);
    return naturalCompare(key1, key2);
  }));
}

export { lexicographicSortSchema };
