import { visitInParallel } from "../language/visitor.mjs";

import { assertValidSchema } from "../type/validate.mjs";

import { visitWithTypeInfo, TypeInfo } from "../utilities/TypeInfo.mjs";

import { specifiedRules, specifiedSDLRules } from "./specifiedRules.mjs";

import { ValidationContext, SDLValidationContext } from "./ValidationContext.mjs";

import { GraphQLError, visit } from "../shared.mjs";

function validate(schema, documentAST, rules = specifiedRules, options, typeInfo = new TypeInfo(schema)) {
  var _options$maxErrors;
  const maxErrors = null !== (_options$maxErrors = null == options ? void 0 : options.maxErrors) && void 0 !== _options$maxErrors ? _options$maxErrors : 100;
  assertValidSchema(schema);
  const abortObj = {};
  const errors = [];
  const context = new ValidationContext(schema, documentAST, typeInfo, (error => {
    if (errors.length >= maxErrors) {
      errors.push(new GraphQLError("Too many validation errors, error limit reached. Validation aborted."));
      throw abortObj;
    }
    errors.push(error);
  }));
  const visitor = visitInParallel(rules.map((rule => rule(context))));
  try {
    visit(documentAST, visitWithTypeInfo(typeInfo, visitor));
  } catch (e) {
    if (e !== abortObj) {
      throw e;
    }
  }
  return errors;
}

function validateSDL(documentAST, schemaToExtend, rules = specifiedSDLRules) {
  const errors = [];
  const context = new SDLValidationContext(documentAST, schemaToExtend, (error => {
    errors.push(error);
  }));
  const visitors = rules.map((rule => rule(context)));
  visit(documentAST, visitInParallel(visitors));
  return errors;
}

function assertValidSDL(documentAST) {
  const errors = validateSDL(documentAST);
  if (0 !== errors.length) {
    throw new Error(errors.map((error => error.message)).join("\n\n"));
  }
}

function assertValidSDLExtension(documentAST, schema) {
  const errors = validateSDL(documentAST, schema);
  if (0 !== errors.length) {
    throw new Error(errors.map((error => error.message)).join("\n\n"));
  }
}

export { assertValidSDL, assertValidSDLExtension, validate };
